#!/usr/bin/env python

"""
Test the  WBEMConnection iteration methods (Iter...) by mocking the lower level
WBEMConnection methods that are called by the iteration methods.

Thus, to mock iterEnumerateInstances, we must mock OpenEnumerateInstances,
pullInstanceseWithPath, CloseEnumeration, and EnumerateInstances to be complete.
With that we can execute the variations on IterEnumerateInstances.

This test uses pytest as the test tool.

It specifically tests:
1. To assure that use_pull_operations works
2. The correct parameters are passed to the lower level methods.
3. The correct results are returned.

It does not test for valid parameters if the parameters are tested by lower
level methods because those are the methods that are mocked.
"""

from __future__ import absolute_import, print_function

import pytest
import six

from mock import Mock

from pywbem import WBEMConnection, CIMInstance, CIMClass, CIMInstanceName, \
    CIMProperty, CIMError, CIM_ERR_NOT_SUPPORTED
from pywbem.config import DEFAULT_ITER_MAXOBJECTCOUNT
from pywbem.cim_operations import pull_inst_result_tuple, \
    pull_path_result_tuple, pull_query_result_tuple

# TODO: ks Oct 17 Remove this pylint disable in the future when pylint or
# the pylint plugin for pylint is fixed
# Today it reports error "redefining name ..."
# pylint: disable=redefined-outer-name


@pytest.fixture
def tst_insts():
    """
    Pytest fixture to return an instance
    """
    i = CIMInstance('CIM_Foo',
                    properties={'Name': 'Foo', 'Chicken': 'Ham'},
                    path=CIMInstanceName('CIM_Foo', {'Name': 'Foo'}))
    return[i]


@pytest.fixture
def tst_class():
    """
    Pytest fixture to return an instance
    """
    cl = CIMClass('CIM_Foo', properties={'InstanceID':
                                         CIMProperty('InstanceID', None,
                                                     type='string')})
    return cl


@pytest.fixture
def tst_paths():
    """ Pytest fixture to return an instance name. This creates a list
        of CIMInstanceName objects and returns the list.
    """
    inst_count = 1
    rtn = []
    for i in six.moves.range(inst_count):
        i = CIMInstanceName('CIM_Foo',
                            keybindings={'InstanceID': str(i + 1000)},
                            host='woot.com',
                            namespace='root/cimv2')
        rtn.append(i)
    return rtn

########################################################################
#
#               IterEnumerateInstances tests
#
########################################################################


class TestIterEnumerateInstances(object):
    """Test IterEnumerateInstance Execution"""
    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "lo, di, iq, ico", [[None, None, None, None],
                            [False, False, False, False],
                            [True, True, True, True]]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    def test_orig_operation_success(self, use_pull_param, tst_insts, ns,
                                    di, iq, lo, ico, pl,):
        # pylint: disable=no-self-use
        """
            Test Use of IterEnumerateInstances from IterEnumerateInstances.
            This forces the enumerate by mocking NOT_Supported on the
            OpenEnumerateInstancess.
            It is parameterized to test variations on all parameters. Note
            that it only tests legal parameters.

            It confirms that EnumerateInstances receives the correct parameter
            for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.EnumerateInstances = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_inst_pull_operations == use_pull_param)

        result = [inst for inst in
                  conn.IterEnumerateInstances('CIM_Foo',
                                              namespace=ns,
                                              LocalOnly=lo,
                                              DeepInheritance=di,
                                              IncludeQualifiers=iq,
                                              IncludeClassOrigin=ico,
                                              PropertyList=pl)]

        conn.EnumerateInstances.assert_called_with('CIM_Foo',
                                                   namespace=ns,
                                                   LocalOnly=lo,
                                                   DeepInheritance=di,
                                                   IncludeQualifiers=iq,
                                                   IncludeClassOrigin=ico,
                                                   PropertyList=pl)

        assert(conn._use_enum_inst_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use
        """
            Test operation with use_pull_operation=True and
            pull operation returns exception.
            This must fail since we force use of the pull operations but
            they return error.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.EnumerateInstances = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_enum_inst_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=unused-argument
            _ = [i for i in conn.IterEnumerateInstances('CIM_Foo')]  # noqa=F841

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_enum_inst_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "lo, di, iq, ico", [[None, None, None, None],
                            [False, False, False, False],
                            [True, True, True, True]]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    lo, di, iq, ico, pl,
                                    fl, fq, ot, coe, moc):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This tests with variations of all of the
            input parameters from None to other valid values.  It does not
            do any invalid values.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_inst_pull_operations == use_pull_param)

        result = [inst for inst in
                  conn.IterEnumerateInstances('CIM_Foo',
                                              namespace=ns,
                                              LocalOnly=lo,
                                              DeepInheritance=di,
                                              IncludeQualifiers=iq,
                                              IncludeClassOrigin=ico,
                                              PropertyList=pl,
                                              FilterQueryLanguage=fl,
                                              FilterQuery=fq,
                                              OperationTimeout=ot,
                                              ContinueOnError=coe,
                                              MaxObjectCount=moc)]
        conn.OpenEnumerateInstances.assert_called_with(
            'CIM_Foo',
            namespace=ns,
            LocalOnly=lo,
            DeepInheritance=di,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        # pylint: disable=protected-access
        assert(conn._use_enum_inst_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=[], eos=False,
                                                     context='blah'))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_inst_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterEnumerateInstances('CIM_Foo')]

        assert(conn._use_enum_inst_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_insts, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid maxObjectCount input parameter
        """
        conn = WBEMConnection('dummy')

        conn.EnumerateInstances = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            result = [inst for inst in  # noqa F841
                      conn.IterEnumerateInstances('CIM_Foo',
                                                  MaxObjectCount=max_cnt)]


########################################################################
#
#               IterEnumerateInstancePaths tests
#
########################################################################


class TestIterEnumerateInstancePaths(object):
    """Test IterEnumerateInstanceNames execution"""
    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    def test_orig_operation_success(self, use_pull_param, tst_paths, ns):
        # pylint: disable=no-self-use
        """
            Test call with no forcing
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstanceNames = Mock(return_value=tst_paths)
        conn.OpenEnumerateInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterEnumerateInstancePaths(
            'CIM_Foo',
            namespace=ns)]

        conn.EnumerateInstanceNames.assert_called_with('CIM_Foo',
                                                       namespace=ns)

        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use
        """
            Test IterEnumerateInstances with forced pull operation and
            pull operation returns exception
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.EnumerateInstanceNames = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=expression-not-assigned
            [i for i in conn.IterEnumerateInstancePaths('CIM_Foo')]  # noqa=F841

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, tst_paths, ns, fl, fq, ot,
                                    coe, moc, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This runs parametrized tests with all of the
            input parameters for legal values.  It does not test for illegal
            values.
            It all tests the mocs OpenEnumerateInstances for correct
            parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstanceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations == use_pull_param)

        result = [p for p in conn.IterEnumerateInstancePaths(
            'CIM_Foo',
            namespace=ns,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)]

        conn.OpenEnumerateInstancePaths.assert_called_with(
            'CIM_Foo',
            namespace=ns,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_paths, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstanceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=[], eos=False,
                                                     context='blah'))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterEnumerateInstancePaths('CIM_Foo')]

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_enum_path_pull_operations is True)

        assert(result == tst_paths)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_paths, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid max_object_count
        """
        conn = WBEMConnection('dummy')

        conn.EnumerateInstanceNames = Mock(return_value=tst_paths)
        conn.OpenEnumeratePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            result = [inst for inst in  # noqa F841
                      conn.IterEnumerateInstancePaths('CIM_Foo',
                                                      MaxObjectCount=max_cnt)]


########################################################################
#
#               IterReferenceInstances tests
#
########################################################################


class TestIterReferenceInstances(object):
    """Test IterReferenceInstance Execution"""

    @staticmethod
    def target_path(ns):
        """
            Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    def test_orig_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, iq, ico, pl,):
        # pylint: disable=no-self-use
        """
            Test Use of IterReferenceInstances from IterReferenceInstances.
            This forces the enumerate by mocking NOT_Supported on the
            OpenReferenceInstancess.
            It is parameterized to test variations on all parameters. Note
            that it only tests legal parameters.

            It confirms that ReferenceInstances receives the correct parameter
            for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_ref_inst_pull_operations == use_pull_param)

        result = [inst for inst in
                  conn.IterReferenceInstances(self.target_path(ns),
                                              ResultClass=rc,
                                              Role=ro,
                                              IncludeQualifiers=iq,
                                              IncludeClassOrigin=ico,
                                              PropertyList=pl)]

        conn.References.assert_called_with(self.target_path(ns),
                                           ResultClass=rc,
                                           Role=ro,
                                           IncludeQualifiers=iq,
                                           IncludeClassOrigin=ico,
                                           PropertyList=pl)

        assert(conn._use_ref_inst_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use
        """
            Test operation with use_pull_operation=True and
            pull operation returns exception.
            This must fail since we force use of the pull operations but
            they return error.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_ref_inst_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=unused-argument
            _ = [i for i in conn.IterReferenceInstances(  # noqa=F841
                self.target_path('root/cimv2'))]

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_ref_inst_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, iq, ico, pl, fl, fq, ot, coe,
                                    moc):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This tests with variations of all of the
            input parameters from None to other valid values.  It does not
            do any invalid values.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.References = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_ref_inst_pull_operations == use_pull_param)
        result = [inst for inst in
                  conn.IterReferenceInstances(self.target_path(ns),
                                              ResultClass=rc,
                                              Role=ro,
                                              IncludeQualifiers=iq,
                                              IncludeClassOrigin=ico,
                                              PropertyList=pl,
                                              FilterQueryLanguage=fl,
                                              FilterQuery=fq,
                                              OperationTimeout=ot,
                                              ContinueOnError=coe,
                                              MaxObjectCount=moc)]
        conn.OpenReferenceInstances.assert_called_with(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        # pylint: disable=protected-access
        assert(conn._use_ref_inst_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos. This does not retest the
            variations of input parameters because it is only confirming
            that the pull executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.References = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=[], eos=False,
                                                     context='blah'))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_ref_inst_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterReferenceInstances(
            self.target_path('root/cimv2'))]

        assert(conn._use_ref_inst_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_insts, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid maxObjectCount input parameter
        """
        conn = WBEMConnection('dummy')

        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            result = [inst for inst in  # noqa F841
                      conn.IterReferenceInstances(self.target_path('root/x'),
                                                  MaxObjectCount=max_cnt)]


########################################################################
#
#               IterReferenceInstancePaths tests
#
########################################################################

class TestIterReferenceInstancePaths(object):
    """Test IterReferenceInstancePaths Execution"""

    @staticmethod
    def target_path(ns):
        """
            Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    def test_orig_operation_success(self, use_pull_param, tst_insts, ns, rc,
                                    ro):
        # pylint: disable=no-self-use
        """
            Test Use of IterReferenceInstancePaths from
            IterReferenceInstancePaths.
            This forces the enumerate by mocking NOT_Supported on the
            OpenReferenceInstances.
            It is parameterized to test variations on all parameters. Note
            that it only tests legal parameters.

            It confirms that ReferenceInstances receives the correct parameter
            for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_ref_path_pull_operations == use_pull_param)

        result = [inst for inst in
                  conn.IterReferenceInstancePaths(self.target_path(ns),
                                                  ResultClass=rc,
                                                  Role=ro)]

        conn.ReferenceNames.assert_called_with(self.target_path(ns),
                                               ResultClass=rc,
                                               Role=ro)

        assert(conn._use_ref_path_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use
        """
            Test operation with use_pull_operation=True and
            pull operation returns exception.
            This must fail since we force use of the pull operations but
            they return error.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_ref_path_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=unused-argument
            _ = [i for i in conn.IterReferenceInstancePaths(  # noqa=F841
                self.target_path('root/cimv2'))]

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_ref_path_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_paths, ns,
                                    rc, ro, fl, fq, ot, coe, moc):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This tests with variations of all of the
            input parameters from None to other valid values.  It does not
            do any invalid values.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.ReferenceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_ref_path_pull_operations == use_pull_param)
        result = [inst for inst in
                  conn.IterReferenceInstancePaths(self.target_path(ns),
                                                  ResultClass=rc,
                                                  Role=ro,
                                                  FilterQueryLanguage=fl,
                                                  FilterQuery=fq,
                                                  OperationTimeout=ot,
                                                  ContinueOnError=coe,
                                                  MaxObjectCount=moc)]
        conn.OpenReferenceInstancePaths.assert_called_with(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        # pylint: disable=protected-access
        assert(conn._use_ref_path_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_paths, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos. This does not retest the
            variations of input parameters because it is only confirming
            that the pull executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.ReferenceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=[], eos=False,
                                                     context='blah'))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_ref_path_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterReferenceInstancePaths(
            self.target_path('root/cimv2'))]

        assert(conn._use_ref_path_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_insts, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid maxObjectCount input parameter
        """
        conn = WBEMConnection('dummy')

        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            result = [inst for inst in  # noqa F841
                      conn.IterReferenceInstancePaths(self.target_path('dum'),
                                                      MaxObjectCount=max_cnt)]

########################################################################
#
#               IterAssociatorInstances tests
#
########################################################################


class TestIterAssociatorInstances(object):
    """Test IterReferenceInstance Execution"""

    @staticmethod
    def target_path(ns):
        """
            Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    def test_orig_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, ac, rr, iq, ico, pl,):
        # pylint: disable=no-self-use
        """
            Test Use of IterAssociatorInstances from IterAssociatorInstances.
            This forces the enumerate by mocking NOT_Supported on the
            OpenAssociatorInstancess.
            It is parameterized to test variations on all parameters. Note
            that it only tests legal parameters.

            It confirms that AssociatorInstances receives the correct parameter
            for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_inst_pull_operations == use_pull_param)

        result = [inst for inst in
                  conn.IterAssociatorInstances(self.target_path(ns),
                                               AssocClass=ac,
                                               ResultClass=rc,
                                               Role=ro,
                                               ResultRole=rr,
                                               IncludeQualifiers=iq,
                                               IncludeClassOrigin=ico,
                                               PropertyList=pl)]

        conn.Associators.assert_called_with(self.target_path(ns),
                                            AssocClass=ac,
                                            ResultClass=rc,
                                            Role=ro,
                                            ResultRole=rr,
                                            IncludeQualifiers=iq,
                                            IncludeClassOrigin=ico,
                                            PropertyList=pl)

        assert(conn._use_assoc_inst_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use
        """
            Test operation with use_pull_operation=True and
            pull operation returns exception.
            This must fail since we force use of the pull operations but
            they return error.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_assoc_inst_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=unused-argument
            _ = [i for i in conn.IterAssociatorInstances(  # noqa=F841
                self.target_path('root/cimv2'))]

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_assoc_inst_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, rr, ac, iq, ico, pl, fl, fq, ot,
                                    coe, moc):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This tests with variations of all of the
            input parameters from None to other valid values.  It does not
            do any invalid values.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.Associators = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_inst_pull_operations == use_pull_param)
        result = [inst for inst in
                  conn.IterAssociatorInstances(self.target_path(ns),
                                               AssocClass=ac,
                                               ResultClass=rc,
                                               Role=ro,
                                               ResultRole=rr,
                                               IncludeQualifiers=iq,
                                               IncludeClassOrigin=ico,
                                               PropertyList=pl,
                                               FilterQueryLanguage=fl,
                                               FilterQuery=fq,
                                               OperationTimeout=ot,
                                               ContinueOnError=coe,
                                               MaxObjectCount=moc)]
        conn.OpenAssociatorInstances.assert_called_with(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        # pylint: disable=protected-access
        assert(conn._use_assoc_inst_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos. This does not retest the
            variations of input parameters because it is only confirming
            that the pull executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.Associators = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=[], eos=False,
                                                     context='blah'))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_inst_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterAssociatorInstances(
            self.target_path('root/cimv2'))]

        assert(conn._use_assoc_inst_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_insts)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_insts, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid maxObjectCount input parameter
        """
        conn = WBEMConnection('dummy')

        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            result = [inst for inst in  # noqa F841
                      conn.IterAssociatorInstances(self.target_path('root/x'),
                                                   MaxObjectCount=max_cnt)]

########################################################################
#
#               IterAssociatorInstancePaths tests
#
########################################################################


class TestIterAssociatorInstancePaths(object):  # pylint: disable=invalid-name
    """Test IterAssociatorInstancePaths Execution."""

    @staticmethod
    def target_path(ns):
        """
            Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    def test_orig_operation_success(self, use_pull_param, tst_paths, ns, rc,
                                    ro, rr, ac):
        # pylint: disable=no-self-use
        """
            Test Use of IterAssociatorInstancePaths from
            IterAssociatorInstancePaths.
            This forces the enumerate by mocking NOT_Supported on the
            OpenReferenceInstances.
            It is parameterized to test variations on all parameters. Note
            that it only tests legal parameters.

            It confirms that ReferenceInstances receives the correct parameter
            for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations == use_pull_param)

        result = [inst for inst in
                  conn.IterAssociatorInstancePaths(self.target_path(ns),
                                                   AssocClass=ac,
                                                   ResultClass=rc,
                                                   Role=ro,
                                                   ResultRole=rr)]

        conn.AssociatorNames.assert_called_with(self.target_path(ns),
                                                AssocClass=ac,
                                                ResultClass=rc,
                                                Role=ro,
                                                ResultRole=rr)

        assert(conn._use_assoc_path_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    def test_operation_fail(self, tst_paths):
        # pylint: disable=no-self-use
        """
            Test operation with use_pull_operation=True and
            pull operation returns exception.
            This must fail since we force use of the pull operations but
            they return error.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=unused-argument
            _ = [i for i in conn.IterAssociatorInstancePaths(  # noqa=F841
                self.target_path('root/cimv2'))]

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_paths, ns,
                                    rc, ro, rr, ac, fl, fq, ot, coe, moc):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This tests with variations of all of the
            input parameters from None to other valid values.  It does not
            do any invalid values.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.AssociatorNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations == use_pull_param)
        result = [inst for inst in
                  conn.IterAssociatorInstancePaths(self.target_path(ns),
                                                   AssocClass=ac,
                                                   ResultClass=rc,
                                                   Role=ro,
                                                   ResultRole=rr,
                                                   FilterQueryLanguage=fl,
                                                   FilterQuery=fq,
                                                   OperationTimeout=ot,
                                                   ContinueOnError=coe,
                                                   MaxObjectCount=moc)]
        conn.OpenAssociatorInstancePaths.assert_called_with(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_paths, use_pull_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos. This does not retest the
            variations of input parameters because it is only confirming
            that the pull executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.AssociatorNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=[], eos=False,
                                                     context='blah'))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations == use_pull_param)

        result = [inst for inst in conn.IterAssociatorInstancePaths(
            self.target_path('root/cimv2'))]

        assert(conn._use_assoc_path_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result == tst_paths)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_paths, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid maxObjectCount input parameter
        """
        conn = WBEMConnection('dummy')

        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            result = [inst for inst in  # noqa F841
                      conn.IterAssociatorInstancePaths(self.target_path('dum'),
                                                       MaxObjectCount=max_cnt)]


########################################################################
#
#               IterQueryInstances tests
#
########################################################################


class TestIterQueryInstances(object):  # pylint: disable=invalid-name
    """Test IterAssociatorInstancePaths Execution."""

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "ql, query", [(None, None), ('CQL', 'SELECT from *')]
    )
    def test_orig_operation_success(self, use_pull_param, tst_insts, ns,
                                    ql, query):
        # pylint: disable=no-self-use
        """
            Test Use IterQueryInstances using the original ExecQuery op
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.ExecQuery = Mock(return_value=tst_insts)
        conn.OpenQueryInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_query_pull_operations == use_pull_param)

        q_result = conn.IterQueryInstances(ql, query, namespace=ns)
        result_insts = [inst for inst in q_result.generator]

        conn.ExecQuery.assert_called_with(ql, query, namespace=ns)

        assert(q_result.query_result_class is None)
        assert(conn._use_query_pull_operations is False)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result_insts == tst_insts)

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use
        """
            Test operation with use_pull_operation=True and
            pull operation returns exception.
            This must fail since we force use of the pull operations but
            they return error.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.ExecQuery = Mock(return_value=tst_insts)
        conn.OpenQueryInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert(conn.use_pull_operations is True)
        # pylint: disable=protected-access
        assert(conn._use_query_pull_operations is True)

        with pytest.raises(CIMError) as exec_info:
            # pylint: disable=unused-variable
            q_result = conn.IterQueryInstances(  # noqa=F841
                'CQL', 'Select from *')

        exc = exec_info.value
        assert(exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED')

        # pylint: disable=protected-access
        assert(conn._use_query_pull_operations is True)
        assert(conn.use_pull_operations is True)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "ql, query", [(None, None), ('CQL', 'SELECT from *')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    # ReturnQueryResultClass (Boolean flag))
    @pytest.mark.parametrize(
        "rqrc_param", [None, True]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    ql, query, ot, coe, moc, rqrc_param,
                                    tst_class):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation and pull operation
            returns exception. This tests with variations of all of the
            input parameters from None to other valid values.  It does not
            do any invalid values.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # Get class definition for response if ReturnQueryResultClass
        qrc = tst_class if rqrc_param is True else None

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=qrc))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_query_pull_operations == use_pull_param)

        q_result = conn.IterQueryInstances(ql, query, namespace=ns,
                                           ReturnQueryResultClass=rqrc_param,
                                           OperationTimeout=ot,
                                           ContinueOnError=coe,
                                           MaxObjectCount=moc)

        conn.OpenQueryInstances.assert_called_with(
            ql, query, namespace=ns,
            ReturnQueryResultClass=rqrc_param,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        result_insts = [inst for inst in q_result.generator]

        assert(q_result.query_result_class == qrc)

        # pylint: disable=protected-access
        assert(conn._use_query_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result_insts == tst_insts)

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    # ReturnQueryResultClass
    @pytest.mark.parametrize(
        "rqrc_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param,
                                    rqrc_param):
        # pylint: disable=no-self-use
        """
            Test call with forced pull operation Open returns nothing and
            pull returns the instance and eos. This does not retest the
            variations of input parameters because it is only confirming
            that the pull executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # Get class definition for response if ReturnQueryResultClass
        qrc = tst_class if rqrc_param is True else None

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(
                instances=[],
                eos=False,
                context="BlahBlah",
                query_result_class=qrc))
        conn.PullInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=qrc))

        assert(conn.use_pull_operations == use_pull_param)
        # pylint: disable=protected-access
        assert(conn._use_assoc_path_pull_operations == use_pull_param)

        q_result = conn.IterQueryInstances('CQL', 'Select from *',
                                           ReturnQueryResultClass=rqrc_param)

        result_insts = [inst for inst in q_result.generator]

        conn.OpenQueryInstances.assert_called_with(
            'CQL', 'Select from *',
            namespace=None,
            ReturnQueryResultClass=rqrc_param,
            OperationTimeout=None,
            ContinueOnError=None,
            MaxObjectCount=DEFAULT_ITER_MAXOBJECTCOUNT)

        # TODO ks: This assert disabled.
        # assert(q_result.query_result_class == rc)

        # pylint: disable=protected-access
        assert(conn._use_query_pull_operations is True)
        assert(conn.use_pull_operations == use_pull_param)
        assert(result_insts == tst_insts)

    @pytest.mark.parametrize(
        "max_cnt", [0, -1]
    )
    def test_invalid_params(self, tst_insts, max_cnt):
        # pylint: disable=no-self-use
        """
        Test for invalid maxObjectCount input parameter
        """
        conn = WBEMConnection('dummy')

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=None))
        with pytest.raises(ValueError):
            # pylint: disable=unused-variable
            q_result = conn.IterQueryInstances(  # noqa=F841
                'CQL', 'Select from *', MaxObjectCount=max_cnt)



#
# (C) Copyright 2003,2004 Hewlett-Packard Development Company, L.P.
# (C) Copyright 2006,2007 Novell, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
# Author: Bart Whiteley <bwhiteley@suse.de>
# Author: Ross Peoples <ross.peoples@gmail.com>
#

# pylint: disable=line-too-long
"""
Python classes for representing values of CIM data types, and related
conversion functions.

The following table shows how CIM data types are represented in Python.
Note that some basic CIM data types are represented with built-in Python
types.

========================================  =====================================
CIM data type                             Python type
========================================  =====================================
boolean                                   :class:`py:bool`
char16                                    :term:`string`
string                                    :term:`string`
string (EmbeddedInstance)                 :class:`~pywbem.CIMInstance`
string (EmbeddedObject)                   :class:`~pywbem.CIMInstance`
                                          or :class:`~pywbem.CIMClass`
datetime                                  :class:`~pywbem.CIMDateTime`
reference                                 :class:`~pywbem.CIMInstanceName`
uint8                                     :class:`~pywbem.Uint8`
uint16                                    :class:`~pywbem.Uint16`
uint32                                    :class:`~pywbem.Uint32`
uint64                                    :class:`~pywbem.Uint64`
sint8                                     :class:`~pywbem.Sint8`
sint16                                    :class:`~pywbem.Sint16`
sint32                                    :class:`~pywbem.Sint32`
sint64                                    :class:`~pywbem.Sint64`
real32                                    :class:`~pywbem.Real32`
real64                                    :class:`~pywbem.Real64`
[] (array)                                :class:`py:list`
========================================  =====================================

The CIM NULL value is represented with Python `None` which can be used for any
CIM typed value to represent NULL.

Note that init methods of pywbem classes that take CIM typed values as input
may support Python types in addition to those shown above. For example, the
:class:`~pywbem.CIMProperty` class represents property values of CIM datetime
type internally as :class:`~pywbem.CIMDateTime` objects, but its init method
accepts :class:`py:datetime.timedelta` objects, :class:`py:datetime.datetime`
objects, :term:`string`, in addition to
:class:`~pywbem.CIMDateTime` objects.
"""
# pylint: enable=line-too-long
# Note: When used before module docstrings, Pylint scopes the disable statement
#       to the whole rest of the file, so we need an enable statement.

# This module is meant to be safe for 'import *'.

from __future__ import absolute_import

from datetime import tzinfo, datetime, timedelta
import re
import warnings
import copy
import traceback
import six

from .config import DEBUG_WARNING_ORIGIN, ENFORCE_INTEGER_RANGE
from ._utils import _ensure_unicode, _hash_item, _format, _to_unicode

if six.PY2:
    # pylint: disable=invalid-name,undefined-variable
    _Longint = long  # noqa: F821
else:
    # pylint: disable=invalid-name
    _Longint = int


__all__ = ['cimtype', 'type_from_name', 'MinutesFromUTC', 'CIMType',
           'CIMDateTime', 'CIMInt', 'Uint8', 'Sint8', 'Uint16', 'Sint16',
           'Uint32', 'Sint32', 'Uint64', 'Sint64', 'CIMFloat', 'Real32',
           'Real64']


class _CIMComparisonMixin(object):  # pylint: disable=too-few-public-methods
    """Mixin class providing default implementations for rich comparison
    operators.

    In Python 2, the rich comparison operators (e.g. `__eq__()`) have
    precedence over the traditional comparator method (`_cmp__()`).
    In Python 3, the comparator method (`_cmp__()`) no longer exists.
    Therefore, implementing the rich comparison operators works in both.

    The default implementations delegate to a comparator method `_cmp()`
    implemented by subclasses. This requires that the subclasses can
    define total ordering. (If they cannot, this mixin class cannot be
    used).
    """

    def __eq__(self, other):
        """
        Invoked when two CIM objects are compared with the `==` operator.

        The comparison is delegated to the `_cmp()` method.
        """
        return self._cmp(other) == 0

    def __ne__(self, other):
        """
        Invoked when two CIM objects are compared with the `!=` operator.

        The comparison is delegated to the `_cmp()` method.
        """
        return self._cmp(other) != 0

    def __ordering_deprecated(self):
        """Deprecated warning for pywbem CIM Objects"""
        msg = _format("Ordering comparisons involving {0} objects are "
                      "deprecated.", self.__class__.__name__)
        if DEBUG_WARNING_ORIGIN:
            msg += "\nTraceback:\n" + ''.join(traceback.format_stack())
        warnings.warn(msg, DeprecationWarning, stacklevel=3)

    def __lt__(self, other):
        """
        Invoked when two CIM objects are compared with the `<` operator.

        The comparison is delegated to the `_cmp()` method.
        """
        self.__ordering_deprecated()
        return self._cmp(other) < 0

    def __gt__(self, other):
        """
        Invoked when two CIM objects are compared with the `>` operator.

        The comparison is delegated to the `_cmp()` method.
        """
        self.__ordering_deprecated()
        return self._cmp(other) > 0

    def __le__(self, other):
        """
        Invoked when two CIM objects are compared with the `<=` operator.

        The comparison is delegated to the `_cmp()` method.
        """
        self.__ordering_deprecated()
        return self._cmp(other) <= 0

    def __ge__(self, other):
        """
        Invoked when two CIM objects are compared with the `>=` operator.

        The comparison is delegated to the `_cmp()` method.
        """
        self.__ordering_deprecated()
        return self._cmp(other) >= 0

    def _cmp(self, other):
        """
        Interface definition for comparator method to be provided by
        subclasses, as follows:
        * If self == other, 0 must be returned.
        * If self < other, -1 must be returned.
        * If self > other, +1 must be returned.
        """
        raise NotImplementedError

    def __hash__(self):
        """
        Interface definition for hash function to be provided by subclasses.

        Background: In order to behave as expected in sets and other hash-based
        collections, the hash values of objects must be equal when the objects
        themselves are considered equal. The default hash function for classes
        is based on `id()` and therefore does not satisfy that requirement.

        Therefore, the CIM objects need to implement a hash function that
        satisfies that requirement.
        """
        raise NotImplementedError


class MinutesFromUTC(tzinfo):
    """
    Timezone information (an implementation of :class:`py:datetime.tzinfo`)
    that represents a fixed offset in +/- minutes from UTC and is thus suitable
    for the CIM datetime data type.

    Objects of this class are needed in order to make
    :class:`py:datetime.datetime` objects timezone-aware, in order to be
    useable as input data to the timezone-aware :class:`~pywbem.CIMDateTime`
    type.

    They are also used to provide timezone information to
    :meth:`~pywbem.CIMDateTime.now` and
    :meth:`~pywbem.CIMDateTime.fromtimestamp`

    Example:

    ::

        from datetime import datetime
        from time import time
        import pywbem

        # Create a timezone-aware datetime object (for CEDT = UTC+2h), and
        # convert that to CIM datetime:

        dt = datetime(year=2016, month=3, day=31, hour=19, minute=30,
                      second=40, microsecond=654321,
                      tzinfo=pywbem.MinutesFromUTC(120))
        cim_dt = pywbem.CIMDateTime(dt)

        # Convert a POSIX timestamp value to CIM datetime (for EST = UTC-5h):

        posix_ts = time()  # seconds since the epoch, not timezone-aware
        cim_dt = pywbem.CIMDateTime.fromtimestamp(posix_ts,
                                                  pywbem.MinutesFromUTC(-300))
    """

    def __init__(self, offset):  # pylint: disable=super-init-not-called
        """
        Parameters:

          offset (:term:`integer`):
            Timezone offset to be represented in the CIM datetime value in +/-
            minutes from UTC.

            This is the offset of local time to UTC (including DST offset),
            where a positive value indicates minutes east of UTC, and a
            negative value indicates minutes west of UTC.
        """
        self._offset = offset

    def __repr__(self):
        return _format(
            "MinutesFromUTC("
            "offset={s._offset!A})",
            s=self)

    def utcoffset(self, dt):  # pylint: disable=unused-argument
        """
        An implementation of the corresponding base class method
        (see :meth:`py:datetime.tzinfo.utcoffset` for its description),
        which needs
        to return the offset of local time to UTC (including DST offset), as a
        :class:`py:datetime.timedelta` object. This method is called by the
        Python datetime classes, and a pywbem user normally does not have
        to deal with it.

        This implementation returns the offset used to initialize the object,
        for any specified `dt` parameter.
        """
        return timedelta(minutes=self._offset)

    def dst(self, dt):  # pylint: disable=unused-argument
        """
        An implementation of the corresponding base class method,
        (see :meth:`py:datetime.tzinfo.dst` for its description),
        which needs
        to return the offset caused by DST, as a :class:`py:datetime.timedelta`
        object. This method is called by the Python datetime classes, and a
        pywbem user normally does not have to deal with it.

        This implementation returns an offset of 0 (indicating that DST is not
        in effect), for any specified `dt` parameter, because CIM datetime
        values do not represent DST information.
        """
        return timedelta(0)


class CIMType(object):  # pylint: disable=too-few-public-methods
    """Base type for all CIM data types defined in this package."""

    # Note: __str__() is not needed; the inherited method is used,
    # even though there is a __repr__() method here.

    #: The name of the CIM datatype, as a :term:`string`. See
    #: :ref:`CIM data types` for details.
    cimtype = None

    def __repr__(self):
        """Return a string representation suitable for debugging."""
        return _format(
            "{s.__class__.__name__}("
            "cimtype={s.cimtype!A}, "
            "{s})",
            s=self)


class CIMDateTime(CIMType, _CIMComparisonMixin):
    """
    A value of CIM data type datetime.

    The object represents either a timezone-aware point in time, or a time
    interval.

    Two objects of this class compare equal if their public attributes compare
    equal. Objects of this class are immutable and :term:`hashable`, with the
    hash value being based on their public attributes.
    """

    #: The name of the CIM datatype ``"datetime"``
    cimtype = 'datetime'

    _timestamp_pattern = re.compile(
        r'^([\d\*]{4})([\d\*]{2})([\d\*]{2})'
        r'([\d\*]{2})([\d\*]{2})([\d\*]{2})\.([\d\*]{6})'
        r'([+|-])(\d{3})')

    _interval_pattern = re.compile(
        r'^([\d\*]{8})([\d\*]{2})([\d\*]{2})([\d\*]{2})\.([\d\*]{6})'
        r'(:)(000)')

    def __init__(self, dtarg):
        """
        Parameters:

          dtarg:
            The value from which the object is initialized, as one of the
            following types:

            * A :term:`string` object will be
              interpreted as CIM datetime format (see :term:`DSP0004`) and
              will result in a point in time or a time interval. The use
              of asterisk characters in the value is supported according to
              the rules defined in :term:`DSP0004` (e.g.
              "20180911124613.128***:000").
            * A :class:`py:datetime.datetime` object will result in a point
              in time. If the :class:`py:datetime.datetime` object is
              timezone-aware (see :class:`~pywbem.MinutesFromUTC`), the
              specified timezone will be used. Otherwise, a default timezone
              of UTC will be assumed.
            * A :class:`py:datetime.timedelta` object will result in a time
              interval.
            * Another :class:`~pywbem.CIMDateTime` object will be copied.
        """
        self.__timedelta = None  # timedelta value, if interval
        self.__datetime = None  # datetime value, if timestamp
        self.__precision = None  # 0-based index of first asterisk, or None
        dtarg = _ensure_unicode(dtarg)
        if isinstance(dtarg, six.text_type):
            m = self._timestamp_pattern.search(dtarg)
            if m is not None:
                # timestamp format
                parts = m.groups()
                offset = int(parts[8])
                if parts[7] == '-':
                    offset = -offset

                if '*' in dtarg:
                    first = dtarg.index('*')
                    after = dtarg.rindex('*') + 1
                    if not re.match(r'^[\*\.]+$', dtarg[first:after]):
                        raise ValueError(
                            _format("Asterisks in CIM datetime timestamp "
                                    "value are not consecutive: {0!A}", dtarg))
                    if after != 21:  # end of microseconds field
                        raise ValueError(
                            _format("Asterisks in CIM datetime timestamp "
                                    "value do not include least significant "
                                    "field: {0!A}", dtarg))
                    self.__precision = first

                year = self._to_int(parts[0], 0, None, 'year', dtarg)
                month = self._to_int(parts[1], 1, None, 'month', dtarg)
                day = self._to_int(parts[2], 1, None, 'day', dtarg)
                hour = self._to_int(parts[3], 0, None, 'hour', dtarg)
                minute = self._to_int(parts[4], 0, None, 'minute', dtarg)
                second = self._to_int(parts[5], 0, None, 'second', dtarg)
                microsec = self._to_int(parts[6], 0, '0', 'microsecond', dtarg)

                try:
                    # Possible errors are e.g. field out of range
                    self.__datetime = datetime(
                        year, month, day, hour, minute, second, microsec,
                        MinutesFromUTC(offset))
                except ValueError as exc:
                    raise ValueError(
                        _format("Invalid datetime() input from CIM datetime "
                                "timestamp value {0!A}: {1}", dtarg, exc))

            else:
                m = self._interval_pattern.search(dtarg)
                if m is not None:
                    # interval format
                    parts = m.groups()
                    if '*' in dtarg:
                        first = dtarg.index('*')
                        after = dtarg.rindex('*') + 1
                        if not re.match(r'^[\*\.]+$', dtarg[first:after]):
                            raise ValueError(
                                _format("Asterisks in CIM datetime interval "
                                        "value are not consecutive: {0!A}",
                                        dtarg))
                        if after != 21:  # end of microseconds field
                            raise ValueError(
                                _format("Asterisks in CIM datetime interval "
                                        "value do not include least "
                                        "significant field: {0!A}", dtarg))
                        self.__precision = first
                    days = self._to_int(parts[0], 0, None, 'days', dtarg)
                    hours = self._to_int(parts[1], 0, None, 'hours', dtarg)
                    minutes = self._to_int(parts[2], 0, None, 'minutes', dtarg)
                    seconds = self._to_int(parts[3], 0, None, 'seconds', dtarg)
                    microsecs = self._to_int(parts[4], 0, '0', 'microseconds',
                                             dtarg)

                    try:
                        # Possible errors are e.g. field out of range
                        self.__timedelta = timedelta(
                            days=days, hours=hours, minutes=minutes,
                            seconds=seconds, microseconds=microsecs)
                    except ValueError as exc:
                        raise ValueError(
                            _format("Invalid timedelta() input from CIM "
                                    "datetime interval value {0!A}: {1}",
                                    dtarg, exc))
                else:
                    raise ValueError(
                        _format("Invalid format of CIM datetime value: {0!A}",
                                dtarg))
        elif isinstance(dtarg, datetime):
            if dtarg.tzinfo is None:
                self.__datetime = dtarg.replace(tzinfo=MinutesFromUTC(0))
            else:
                self.__datetime = copy.copy(dtarg)
        elif isinstance(dtarg, timedelta):
            self.__timedelta = copy.copy(dtarg)
        elif isinstance(dtarg, CIMDateTime):
            self.__datetime = copy.copy(dtarg.datetime)
            self.__timedelta = copy.copy(dtarg.timedelta)
        else:
            raise TypeError(
                _format("dtarg argument {0!A} has an invalid type: {1} "
                        "(expected datetime, timedelta, string, or "
                        "CIMDateTime)", dtarg, type(dtarg)))

    @staticmethod
    def _to_int(value_str, min_value, rep_digit, field_name, dtarg):
        """
        Convert value_str into an integer, replacing right-consecutive
        asterisks with rep_digit, and an all-asterisk value with min_value.

        field_name and dtarg are passed only for informational purposes.
        """
        if '*' in value_str:
            first = value_str.index('*')
            after = value_str.rindex('*') + 1
            if value_str[first:after] != '*' * (after - first):
                raise ValueError(
                    _format("Asterisks in {0} field of CIM datetime value "
                            "{1!A} are not consecutive: {2!A}",
                            field_name, dtarg, value_str))
            if after != len(value_str):
                raise ValueError(
                    _format("Asterisks in {0} field of CIM datetime value "
                            "{1!A} do not end at end of field: {2!A}",
                            field_name, dtarg, value_str))
            if rep_digit is None:  # pylint: disable=no-else-return
                # Must be an all-asterisk field
                if first != 0:
                    raise ValueError(
                        _format("Asterisks in {0} field of CIM datetime value "
                                "{1!A} do not start at begin of field: {2!A}",
                                field_name, dtarg, value_str))
                return min_value
            else:
                value_str = value_str.replace('*', rep_digit)
        # Because the pattern and the asterisk replacement mechanism already
        # ensure only decimal digits, we expect the integer conversion to
        # always succeed.
        value = int(value_str)
        return value

    def _to_str(self, value, field_begin, field_len):
        """
        Convert value (int) to a field string, considering precision.
        """
        value_str = '{0:0{1}d}'.format(value, field_len)
        if self.precision is not None and \
                self.precision < field_begin + field_len:
            # field is partly or completely affected by precision
            # -> replace insignificant digits with asterisks
            precision_index = max(0, self.precision - field_begin)
            value_str = value_str[:precision_index] + \
                '*' * (field_len - precision_index)
        return value_str

    @property
    def minutes_from_utc(self):
        """
        The timezone offset of this point in time object as +/- minutes from
        UTC.

        A positive value of the timezone offset indicates minutes east of UTC,
        and a negative value indicates minutes west of UTC.

        0, if this object represents a time interval.
        """
        offset = 0
        if self.__datetime is not None and \
                self.__datetime.utcoffset() is not None:
            offset = self.__datetime.utcoffset().seconds / 60
            if self.__datetime.utcoffset().days == -1:
                offset = -((60 * 24) - offset)
        return int(offset)

    @property
    def datetime(self):
        """
        The point in time represented by this object, as a
        :class:`py:datetime.datetime` object.

        `None` if this object represents a time interval.
        """
        return self.__datetime

    @property
    def timedelta(self):
        """
        The time interval represented by this object, as a
        :class:`py:datetime.timedelta` object.

        `None` if this object represents a point in time.
        """
        return self.__timedelta

    @property
    def precision(self):
        """
        Precision of the time interval or point in time represented by this
        object, if the datetime input string contained asterisk characters.

        The precision is the 0-based index of the first asterisk character in
        the datetime input string, or `None` if there were no asterisk
        characters. For example, the precision of the timestamp value
        "201809121230**.******+000" is 12.
        """
        return self.__precision

    @property
    def is_interval(self):
        """
        A boolean indicating whether this object represents a time interval
        (`True`) or a point in time (`False`).
        """
        return self.__timedelta is not None

    @staticmethod
    def get_local_utcoffset():
        """
        Return the timezone offset of the current local timezone as +/- minutes
        from UTC.

        A positive value indicates minutes east of UTC, and a negative
        value indicates minutes west of UTC.
        """
        utc = datetime.utcnow()
        local = datetime.now()
        if local < utc:
            return - int(float((utc - local).seconds) / 60 + .5)
        return int(float((local - utc).seconds) / 60 + .5)

    @classmethod
    def now(cls, tzi=None):
        """
        Factory method that returns a new :class:`~pywbem.CIMDateTime` object
        representing the current date and time.

        The optional timezone information is used to convert the CIM datetime
        value into the desired timezone. That does not change the point in time
        that is represented by the value, but it changes the value of the
        ``hhmmss`` components of the CIM datetime value to compensate for
        changes in the timezone offset component.

        Parameters:

          tzi (:class:`~pywbem.MinutesFromUTC`):
            Timezone information. `None` means that the current local timezone
            is used.

        Returns:

            A new :class:`~pywbem.CIMDateTime` object representing the current
            date and time.
        """
        if tzi is None:
            tzi = MinutesFromUTC(cls.get_local_utcoffset())
        return cls(datetime.now(tzi))

    @classmethod
    def fromtimestamp(cls, ts, tzi=None):
        # pylint: disable=invalid-name
        """
        Factory method that returns a new :class:`~pywbem.CIMDateTime` object
        from a POSIX timestamp value and optional timezone information.

        A POSIX timestamp value is the number of seconds since "the epoch",
        i.e. 1970-01-01 00:00:00 UTC. Thus, a POSIX timestamp value is
        unambiguous w.r.t. the timezone, but it is not timezone-aware.

        The optional timezone information is used to convert the CIM datetime
        value into the desired timezone. That does not change the point in time
        that is represented by the value, but it changes the value of the
        ``hhmmss`` components of the CIM datetime value to compensate for
        changes in the timezone offset component.

        Parameters:

          ts (:term:`integer`):
            POSIX timestamp value.

          tzi (:class:`~pywbem.MinutesFromUTC`):
            Timezone information. `None` means that the current local timezone
            is used.

        Returns:

            A new :class:`~pywbem.CIMDateTime` object representing the
            specified point in time.
        """
        if tzi is None:
            tzi = MinutesFromUTC(cls.get_local_utcoffset())
        return cls(datetime.fromtimestamp(ts, tzi))

    def __str__(self):
        """
        Return a string representing the object in CIM datetime format.
        """
        if self.is_interval:  # pylint: disable=no-else-return
            days = self.timedelta.days
            hours = self.timedelta.seconds // 3600
            sec_in_hour = self.timedelta.seconds - hours * 3600
            minutes = sec_in_hour // 60
            seconds = sec_in_hour - minutes * 60
            microsecs = self.timedelta.microseconds

            days_str = self._to_str(days, 0, 8)
            hours_str = self._to_str(hours, 8, 2)
            minutes_str = self._to_str(minutes, 10, 2)
            seconds_str = self._to_str(seconds, 12, 2)
            microsecs_str = self._to_str(microsecs, 15, 6)

            ret_str = '{0}{1}{2}{3}.{4}:000'.format(
                days_str, hours_str, minutes_str, seconds_str,
                microsecs_str)
            return ret_str

        else:  # timestamp
            offset = self.minutes_from_utc
            sign = '+'
            if offset < 0:
                sign = '-'
                offset = -offset

            year = self.datetime.year
            month = self.datetime.month
            day = self.datetime.day
            hour = self.datetime.hour
            minute = self.datetime.minute
            second = self.datetime.second
            microsec = self.datetime.microsecond

            year_str = self._to_str(year, 0, 4)
            month_str = self._to_str(month, 4, 2)
            day_str = self._to_str(day, 6, 2)
            hour_str = self._to_str(hour, 8, 2)
            minute_str = self._to_str(minute, 10, 2)
            second_str = self._to_str(second, 12, 2)
            microsec_str = self._to_str(microsec, 15, 6)

            ret_str = '{0}{1}{2}{3}{4}{5}.{6}{7}{8:03d}'.format(
                year_str, month_str, day_str, hour_str, minute_str,
                second_str, microsec_str, sign, offset)
            return ret_str

    def __repr__(self):
        """
        Return a string representation suitable for debugging.
        """
        return _format(
            "CIMDateTime("
            "cimtype={s.cimtype!A}, "
            "datetime={s.datetime!A}, "
            "timedelta={s.timedelta!A}, "
            "precision={s.precision!A}), "
            "minutes_from_utc={s.minutes_from_utc!A})",
            s=self)

    def __getstate__(self):
        return str(self)

    def __setstate__(self, arg):
        self.__init__(arg)

    def _cmp(self, other):
        # Defer import due to circular import dependencies:
        from .cim_obj import cmpitem
        if self is other:
            return 0

        if not isinstance(other, CIMDateTime):
            return 1

        return (cmpitem(self.datetime, other.datetime) or
                cmpitem(self.timedelta, other.timedelta))

    def __hash__(self):
        """
        Return a hash value based on the public attributes of this class.
        Because these attributes are not modifiable, objects of this class are
        :term:`hashable` (and not just :term:`unchanged-hashable`).
        """
        hashes = (
            _hash_item(self.datetime),
            _hash_item(self.timedelta),
            # The 'is_interval' and 'minutes_from_utc' attributes are not used
            # for hash value calculation because they are derived attributes.
        )
        return hash(hashes)


# CIM integer types


class CIMInt(CIMType, _Longint):
    """
    Base type for CIM integer data types. Derived from :class:`~pywbem.CIMType`
    and :class:`py:int` (for Python 3) or :class:`py:long` (for Python 2).

    This class has a concept of a valid range for the represented integer,
    based upon the capability of the CIM data type as defined in
    :term:`DSP0004`. The additional constraints defined by possible MinValue
    or MaxValue qualifiers are not taken into account at this level.

    The valid value range is enforced when an instance of a subclass of this
    class (e.g. :class:`~pywbem.Uint8`) is created. Values outside of the
    valid range raise a :exc:`ValueError`.
    The enforcement of the valid value range can be disabled via the
    configuration variable :data:`~pywbem.config.ENFORCE_INTEGER_RANGE`.

    Two objects of subclasses of this base class compare equal if their numeric
    values compare equal. Objects of this class are immutable and
    :term:`hashable`, with the hash value being based on its numeric value.

    Instances of subclasses of this class can be initialized with the usual
    input arguments supported by :term:`integer`, for example:

    ::

        >>> pywbem.Uint8(42)
        Uint8(cimtype='uint8', 42)

        >>> pywbem.Uint8('42')
        Uint8(cimtype='uint8', 42)

        >>> pywbem.Uint8('2A', 16)
        Uint8(cimtype='uint8', 42)

        >>> pywbem.Uint8('100', 16)
        Traceback (most recent call last):
          . . .
        ValueError: Integer value 256 is out of range for CIM datatype uint8

        >>> pywbem.Uint8(100, 10)
        Traceback (most recent call last):
          . . .
        TypeError: int() can't convert non-string with explicit base
    """

    #: The minimum valid value for the integer, according to the capabilities
    #: of its CIM data type. See :ref:`CIM data types` for a list of CIM
    #: integer data types.
    minvalue = None

    #: The maximum valid value for the integer, according to the capabilities
    #: of its CIM data type. See :ref:`CIM data types` for a list of CIM
    #: integer data types.
    maxvalue = None

    def __new__(cls, *args, **kwargs):

        # Python 3.7 removed support for passing the value for int() as a
        # keyword argument named 'x'. It now needs to be passed as a positional
        # argument. The testclient test case definitions rely on a keyword
        # argument, so we now transform the keyword arg into a positional
        # arg.
        if 'x' in kwargs:
            args = list(*args)  # args is passed as a tuple
            args.append(kwargs.pop('x'))

        value = _Longint(*args, **kwargs)
        if ENFORCE_INTEGER_RANGE:
            if value > cls.maxvalue or value < cls.minvalue:
                raise ValueError(
                    _format("Integer value {0} is out of range for CIM "
                            "datatype {1}", value, cls.cimtype))
        # The value needs to be processed here, because int/long is unmutable
        return super(CIMInt, cls).__new__(cls, *args, **kwargs)

    def __repr__(self):
        """
        Return a string representation suitable for debugging.
        """
        return _format(
            "{s.__class__.__name__}("
            "cimtype={s.cimtype!A}, "
            "minvalue={s.minvalue}, "  # Avoid long indicator 'L' in Python 2
            "maxvalue={s.maxvalue}, "  # Avoid long indicator 'L' in Python 2
            "{s})",
            s=self)


class Uint8(CIMInt):
    """
    A value of CIM data type uint8. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'uint8'
    #: The minimum valid value for the CIM datatype
    minvalue = 0
    #: The maximum valid value for the CIM datatype
    maxvalue = 2**8 - 1


class Sint8(CIMInt):
    """
    A value of CIM data type sint8. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'sint8'
    #: The minimum valid value for the CIM datatype
    minvalue = -2 ** (8 - 1)
    #: The maximum valid value for the CIM datatype
    maxvalue = 2 ** (8 - 1) - 1


class Uint16(CIMInt):
    """
    A value of CIM data type uint16. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'uint16'
    #: The minimum valid value for the CIM datatype
    minvalue = 0
    #: The maximum valid value for the CIM datatype
    maxvalue = 2**16 - 1


class Sint16(CIMInt):
    """
    A value of CIM data type sint16. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'sint16'
    #: The minimum valid value for the CIM datatype
    minvalue = -2 ** (16 - 1)
    #: The maximum valid value for the CIM datatype
    maxvalue = 2 ** (16 - 1) - 1


class Uint32(CIMInt):
    """
    A value of CIM data type uint32. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'uint32'
    #: The minimum valid value for the CIM datatype
    minvalue = 0
    #: The maximum valid value for the CIM datatype
    maxvalue = 2 ** 32 - 1


class Sint32(CIMInt):
    """
    A value of CIM data type sint32. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'sint32'
    #: The minimum valid value for the CIM datatype
    minvalue = -2 ** (32 - 1)
    #: The maximum valid value for the CIM datatype
    maxvalue = 2 ** (32 - 1) - 1


class Uint64(CIMInt):
    """
    A value of CIM data type uint64. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'uint64'
    #: The minimum valid value for the CIM datatype
    minvalue = 0
    #: The maximum valid value for the CIM datatype
    maxvalue = 2 ** 64 - 1


class Sint64(CIMInt):
    """
    A value of CIM data type sint64. Derived from :class:`~pywbem.CIMInt`.

    For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
    """
    #: The name of the CIM datatype
    cimtype = 'sint64'
    #: The minimum valid value for the CIM datatype
    minvalue = -2 ** (64 - 1)
    #: The maximum valid value for the CIM datatype
    maxvalue = 2 ** (64 - 1) - 1


# CIM float types


class CIMFloat(CIMType, float):
    """
    Base type for real (floating point) CIM data types.

    Two objects of subclasses of this base class compare equal if their numeric
    values compare equal. Objects of this class are immutable and
    :term:`hashable`, with the hash value being based on its numeric value.

    Note that equality comparison of floating point numbers in Python (and in
    almost any programming language) comes with some surprises.
    See `"Floating Point Arithmetic: Issues and Limitations"
    <https://docs.python.org/2/tutorial/floatingpoint.html>`_ for details,
    and specifically `"Comparing Floating Point Numbers, 2012 Edition"
    <https://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/>`_
    on the topic of equality comparison. The same issues apply to hash values
    that are based on the numeric value of floating point numbers. Therefore,
    it is not recommended to perform equality comparison of objects of
    subclasses of this class, or to use them as dictionary keys or as members
    in sets.
    """


class Real32(CIMFloat):
    """
    A value of CIM data type real32. Derived from :class:`~pywbem.CIMFloat`.

    It is not recommended to perform equality comparison on objects of this
    class, or to use them as dictionary keys or as members in sets. See
    :class:`~pywbem.CIMFloat` for details.
    """
    #: The name of the CIM datatype
    cimtype = 'real32'


class Real64(CIMFloat):
    """
    A value of CIM data type real64. Derived from :class:`~pywbem.CIMFloat`.

    It is not recommended to perform equality comparison on objects of this
    class, or to use them as dictionary keys or as members in sets. See
    :class:`~pywbem.CIMFloat` for details.
    """
    #: The name of the CIM datatype
    cimtype = 'real64'


# Python number types listed in :term:`number`.
number_types = six.integer_types + (float,)  # pylint: disable=invalid-name


def cimtype(obj):
    """
    Return the CIM data type name of a CIM typed object, as a string.

    For an array, the type is determined from the first array element
    (CIM arrays must be homogeneous w.r.t. the type of their elements).
    If the array is empty, that is not possible and
    :exc:`~py:exceptions.ValueError` is raised.

    Note that Python :term:`numbers <number>` are not valid input objects
    because determining their CIM data type (e.g. :class:`~pywbem.Uint8`,
    :class:`~pywbem.Real32`) would require knowing the value range. Therefore,
    :exc:`~py:exceptions.TypeError` is raised in this case.

    Parameters:

      obj (:term:`CIM data type`):
        The object whose CIM data type name is returned.

    Returns:

      :term:`string`: The CIM data type name of the object (e.g. ``"uint8"``).

    Raises:

      TypeError: The object does not have a valid CIM data type.
      ValueError: Cannot determine CIM data type from an empty array.
    """

    if isinstance(obj, CIMType):
        return obj.cimtype

    if isinstance(obj, bool):
        return 'boolean'

    if isinstance(obj, (six.binary_type, six.text_type)):
        # accept both possible types
        return 'string'

    if isinstance(obj, list):
        try:
            obj = obj[0]
        except IndexError:
            raise ValueError("Cannot determine CIM data type from empty array")
        return cimtype(obj)

    if isinstance(obj, (datetime, timedelta)):
        return 'datetime'

    try:
        instancename_type = CIMInstanceName
    except NameError:
        # Defer import due to circular import dependencies:
        from pywbem.cim_obj import CIMInstanceName as instancename_type
    if isinstance(obj, instancename_type):
        return 'reference'

    try:
        instance_type = CIMInstance
    except NameError:
        # Defer import due to circular import dependencies:
        from pywbem.cim_obj import CIMInstance as instance_type
    if isinstance(obj, instance_type):  # embedded instance
        return 'string'

    try:
        class_type = CIMClass
    except NameError:
        # Defer import due to circular import dependencies:
        from pywbem.cim_obj import CIMClass as class_type
    if isinstance(obj, class_type):  # embedded class
        return 'string'

    raise TypeError(
        _format("Object does not have a valid CIM data type: {0!A}", obj))


_TYPE_FROM_NAME = {
    'boolean': bool,
    'string': six.text_type,  # return the preferred type
    'char16': six.text_type,  # return the preferred type
    'datetime': CIMDateTime,
    # 'reference' covered at run time
    'uint8': Uint8,
    'uint16': Uint16,
    'uint32': Uint32,
    'uint64': Uint64,
    'sint8': Sint8,
    'sint16': Sint16,
    'sint32': Sint32,
    'sint64': Sint64,
    'real32': Real32,
    'real64': Real64,
}


def type_from_name(type_name):
    """
    Return the Python type object for a given CIM data type name.

    For example, type name ``"uint8"`` will return type object
    :class:`~pywbem.Uint8`.

    For CIM data type names ``"string"`` and ``"char16"``, the
    :term:`unicode string` type is returned (Unicode strings are the preferred
    representation for these CIM data types).

    The returned type can be used as a constructor from a differently typed
    input value in many cases. Notable limitations are:

    * In Python 3, the :class:`py3:str` type is used to represent CIM string
      data types. When constructing such an object from a byte string, the
      resulting string is not a unicode-translated version of the byte string
      as one would assume (and as is the case in Python 2), but instead that
      results in a unicode string that is a `repr()` representation of the
      byte string::

          string_type = type_from_name('string')  # str
          s1 = b'abc'
          s2 = string_type(s1)  # results in u"b'abc'", and not in u"abc"

      Use `decode()` and `encode()` for strings instead of type conversion
      syntax (in both Python 2 and 3, for consistency).

    Parameters:

      type_name (:term:`string`):
        The simple (=non-array) CIM data type name (e.g. ``"uint8"`` or
        ``"reference"``).

    Returns:

        The Python type object for the CIM data type (e.g.
        :class:`~pywbem.Uint8` or :class:`~pywbem.CIMInstanceName`).

    Raises:

        ValueError: Unknown CIM data type name.
    """
    if type_name == 'reference':
        # Defer import due to circular import dependencies:
        from .cim_obj import CIMInstanceName
        return CIMInstanceName
    try:
        type_obj = _TYPE_FROM_NAME[type_name]
    except KeyError:
        raise ValueError(
            _format("Unknown CIM data type name: {0!A}", type_name))
    return type_obj


def atomic_to_cim_xml(obj):
    """
    Convert an "atomic" scalar value to a CIM-XML string and return that
    string.

    The returned CIM-XML string is ready for use as the text of a CIM-XML
    'VALUE' element.

    Parameters:

      obj (:term:`CIM data type`, :term:`number`, :class:`py:datetime`):
        The "atomic" input value. May be `None`.

        Must not be an array/list/tuple. Must not be a :ref:`CIM object`.

    Returns:

        A :term:`unicode string` object in CIM-XML value format representing
        the input value. `None`, if the input value is `None`.

    Raises:

        TypeError
    """
    if obj is None:  # pylint: disable=no-else-return
        return obj
    elif isinstance(obj, six.text_type):
        return obj
    elif isinstance(obj, six.binary_type):
        return _to_unicode(obj)
    elif isinstance(obj, bool):
        return u'TRUE' if obj else u'FALSE'
    elif isinstance(obj, (CIMInt, six.integer_types, CIMDateTime)):
        return six.text_type(obj)
    elif isinstance(obj, datetime):
        return six.text_type(CIMDateTime(obj))
    elif isinstance(obj, Real32):
        # DSP0201 requirements for representing real32:
        # The significand must be represented with at least 11 digits.
        # The special values must have the case: INF, -INF, NaN.
        s = u'{0:.11G}'.format(obj)
        if s == 'NAN':
            s = u'NaN'
        elif s in ('INF', '-INF'):
            pass
        elif '.' not in s:
            parts = s.split('E')
            parts[0] = parts[0] + '.0'
            s = 'E'.join(parts)
        return s
    elif isinstance(obj, (Real64, float)):
        # DSP0201 requirements for representing real64:
        # The significand must be represented with at least 17 digits.
        # The special values must have the case: INF, -INF, NaN.
        s = u'{0:.17G}'.format(obj)
        if s == 'NAN':
            s = u'NaN'
        elif s in ('INF', '-INF'):
            pass
        elif '.' not in s:
            parts = s.split('E')
            parts[0] = parts[0] + '.0'
            s = 'E'.join(parts)
        return s
    else:
        raise TypeError(
            _format("Value {0!A} has invalid type {1} for conversion to a "
                    "CIM-XML string", obj, type(obj)))



"""
Test file for use with wbemcli -mock-server parameter that adds an instance
to the mock repository
This file assumes that the file simple_mock_model.mof has already been loaded
so the class CIM_Foo exists.
"""

from __future__ import absolute_import, print_function

from pywbem import CIMInstance, CIMInstanceName

_INAME = 'CIM_Foo%s' % 'wbemcli_tst-1'
_INST_PATH = CIMInstanceName('CIM_Foo', {'InstanceID': _INAME})
_INST = CIMInstance('CIM_Foo',
                    properties={'InstanceID': _INAME},
                    path=_INST_PATH)

# CONN is a defined global variable in the wbemcli environment
global CONN  # pylint: disable=global-at-module-level

CONN.add_cimobjects(_INST)  # noqa: F821 pylint: disable=undefined-variable

# test that instance inserted
assert(CONN.GetInstance(_INST_PATH))  # noqa: F821,E501 pylint: disable=undefined-variable



#!/usr/bin/env python

"""
Test CIM types.
"""

from __future__ import absolute_import, print_function

from datetime import timedelta, datetime
import pytest
import six

from pywbem import CIMType, CIMInt, CIMFloat, Uint8, Uint16, Uint32, Uint64, \
    Sint8, Sint16, Sint32, Sint64, Real32, Real64, CIMDateTime, \
    MinutesFromUTC, CIMClass, CIMInstance, CIMInstanceName, CIMClassName, \
    cimtype, type_from_name

from ..utils.pytest_extensions import simplified_test_function


#
# CIM integer data types
#


@pytest.fixture(params=[

    # Each list item is a tuple of:
    # - obj_type (type): CIM integer data type (e.g. Uint8).
    # - exp_cimtype (string): Expected value of cimtype class attribute for
    #   this CIM data type.
    # - exp_minvalue (integer): Expected value of minvalue class attribute for
    #   this CIM data type.
    # - exp_maxvalue (): Expected value of maxvalue class attribute for
    #   this CIM data type.

    # (obj_type, exp_cimtype, exp_minvalue, exp_maxvalue)
    (Uint8, 'uint8', 0, 2**8 - 1),
    (Uint16, 'uint16', 0, 2**16 - 1),
    (Uint32, 'uint32', 0, 2**32 - 1),
    (Uint64, 'uint64', 0, 2**64 - 1),
    (Sint8, 'sint8', -2**7, 2**7 - 1),
    (Sint16, 'sint16', -2**15, 2**15 - 1),
    (Sint32, 'sint32', -2**31, 2**31 - 1),
    (Sint64, 'sint64', -2**63, 2**63 - 1),
], scope='module')
def integer_tuple(request):
    """Utility function to return param from request"""
    return request.param


def test_integer_class_attrs_class(integer_tuple):
    """Test class attrs via class level"""
    # pylint: disable=redefined-outer-name
    obj_type, exp_cimtype, exp_minvalue, exp_maxvalue = integer_tuple
    assert obj_type.cimtype == exp_cimtype
    assert obj_type.minvalue == exp_minvalue
    assert obj_type.maxvalue == exp_maxvalue


def test_integer_class_attrs_inst(integer_tuple):
    """Test class attrs via instance level"""
    # pylint: disable=redefined-outer-name
    obj_type, exp_cimtype, exp_minvalue, exp_maxvalue = integer_tuple
    obj = obj_type(42)
    assert obj.cimtype == exp_cimtype
    assert obj.minvalue == exp_minvalue
    assert obj.maxvalue == exp_maxvalue


def test_integer_inheritance(integer_tuple):
    """Test inheritance"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    obj = obj_type(42)
    assert isinstance(obj, obj_type)
    assert isinstance(obj, CIMType)
    assert isinstance(obj, CIMInt)
    assert not isinstance(obj, CIMFloat)


def test_integer_init_int(integer_tuple):
    """Test initialization from integer value"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    obj = obj_type(42)
    assert obj == 42


def test_integer_init_str(integer_tuple):
    """Test initialization from string value"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    obj = obj_type('42')
    assert obj == 42


def test_integer_init_str_base10(integer_tuple):
    """Test initialization from string value with base 10"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    obj = obj_type('42', 10)
    assert obj == 42


def test_integer_init_str_base16(integer_tuple):
    """Test initialization from string value with base 16"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    obj = obj_type('2A', 16)
    assert obj == 42


def test_integer_init_minimum(integer_tuple):
    """Test initialization from integer value at minimum"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    exp_minvalue = integer_tuple[2]
    obj = obj_type(exp_minvalue)
    assert obj == exp_minvalue


def test_integer_init_maximum(integer_tuple):
    """Test initialization from integer value at maximum"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    exp_maxvalue = integer_tuple[3]
    obj = obj_type(exp_maxvalue)
    assert obj == exp_maxvalue


def test_integer_init_too_low(integer_tuple):
    """Test initialization from integer value below minimum"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    exp_minvalue = integer_tuple[2]
    try:
        obj_type(exp_minvalue - 1)
    except ValueError:
        pass
    else:
        raise AssertionError("ValueError was not raised.")


def test_integer_init_too_high(integer_tuple):
    """Test initialization from integer value above maximum"""
    # pylint: disable=redefined-outer-name
    obj_type = integer_tuple[0]
    exp_maxvalue = integer_tuple[3]
    try:
        obj_type(exp_maxvalue + 1)
    except ValueError:
        pass
    else:
        raise AssertionError("ValueError was not raised.")


#
# CIM real data types
#


@pytest.fixture(params=[

    # Each list item is a tuple of:
    # - obj_type (type): CIM real data type (e.g. Real32).
    # - exp_cimtype (string): Expected value of cimtype class attribute for
    #   this CIM data type.

    # (obj_type, exp_cimtype)
    (Real32, 'real32'),
    (Real64, 'real64'),
], scope='module')
def real_tuple(request):
    """pytest.fixture returns Real types"""
    return request.param


def test_real_class_attrs_class(real_tuple):
    """Test class attrs via class level using real_tuple fixture"""
    # pylint: disable=redefined-outer-name
    obj_type, exp_cimtype = real_tuple
    assert obj_type.cimtype == exp_cimtype


def test_real_class_attrs_inst(real_tuple):
    """Test class attrs via instance level using real_tuple fixture"""
    # pylint: disable=redefined-outer-name
    obj_type, exp_cimtype = real_tuple
    obj = obj_type(42)
    assert obj.cimtype == exp_cimtype


def test_real_inheritance(real_tuple):
    """Test inheritance  using real_tuple fixture"""
    # pylint: disable=redefined-outer-name
    obj_type = real_tuple[0]
    obj = obj_type(42)
    assert isinstance(obj, obj_type)
    assert isinstance(obj, CIMType)
    assert isinstance(obj, CIMFloat)
    assert not isinstance(obj, CIMInt)


def test_real_init_float(real_tuple):
    """Test initialization from floating point value"""
    # pylint: disable=redefined-outer-name
    obj_type = real_tuple[0]
    obj = obj_type(42.0)
    assert obj == 42.0


def test_real_init_str(real_tuple):
    """Test initialization from string value using real_tuple fixture"""
    # pylint: disable=redefined-outer-name
    obj_type = real_tuple[0]
    obj = obj_type('42.0')
    assert obj == 42.0


#
# CIM datetime data type
#


@pytest.fixture(params=[

    # Each list item is a tuple of:
    # (
    #   dtarg: Input argument for CIMDateTime().
    #   exp_kind: Expected kind of datetime: 'interval' or 'timestamp', or
    #             exception type if initialization failed.
    #   exp_datetime: Expected value for datetime property.
    #   exp_timedelta: Expected value for timedelta property.
    #   exp_precision: Expected value for precision property.
    #   exp_minutesfromutc: Expected value for minutes_from_utc property.
    #   exp_str: Expected string value of the CIMDateTime object.
    # )
    (
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        None,
        0,
        '12345678224455.654321:000'
    ),
    (
        '12345678224455.654321:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        None,
        0,
        '12345678224455.654321:000'
    ),
    (
        CIMDateTime('12345678224455.654321:000'),
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        None,
        0,
        '12345678224455.654321:000'
    ),
    (
        '12345678224455.65432*:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654320),
        20,
        0,
        '12345678224455.65432*:000'
    ),
    (
        '12345678224455.6543**:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654300),
        19,
        0,
        '12345678224455.6543**:000'
    ),
    (
        '12345678224455.654***:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654000),
        18,
        0,
        '12345678224455.654***:000'
    ),
    (
        '12345678224455.65****:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=650000),
        17,
        0,
        '12345678224455.65****:000'
    ),
    (
        '12345678224455.6*****:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=600000),
        16,
        0,
        '12345678224455.6*****:000'
    ),
    (
        '12345678224455.******:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=0),
        15,
        0,
        '12345678224455.******:000'
    ),
    (
        '1234567822445*.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '123456782244**.******:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=0,
                  microseconds=0),
        12,
        0,
        '123456782244**.******:000'
    ),
    (
        '12345678224***.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '1234567822****.******:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=0, seconds=0,
                  microseconds=0),
        10,
        0,
        '1234567822****.******:000'
    ),
    (
        '123456782*****.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '12345678******.******:000',
        'interval',
        None,
        timedelta(days=12345678, hours=0, minutes=0, seconds=0,
                  microseconds=0),
        8,
        0,
        '12345678******.******:000'
    ),
    (
        '1234567*******.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '123456********.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '12345*********.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '1234**********.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '123***********.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '12************.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '1*************.******:000',
        ValueError, None, None, None, None, None
    ),
    (
        '**************.******:000',
        'interval',
        None,
        timedelta(days=0, hours=0, minutes=0, seconds=0,
                  microseconds=0),
        0,
        0,
        '**************.******:000'
    ),
    (
        '12345678224455.654321:777',
        ValueError, None, None, None, None, None
    ),
    (
        '12345678224455,654321:000',
        ValueError, None, None, None, None, None
    ),
    (
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        None,
        None,
        120,
        '20140924193040.654321+120'
    ),
    (
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(0)),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(0)),
        None,
        None,
        0,
        '20140924193040.654321+000'
    ),
    (
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(0)),
        None,
        None,
        0,
        '20140924193040.654321+000'
    ),
    (
        '20140924193040.654321+120',
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        None,
        None,
        120,
        '20140924193040.654321+120'
    ),
    (
        '20140924193040.654321+120',
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        None,
        None,
        120,
        '20140924193040.654321+120'
    ),
    (
        '20180912123040.65432*+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=40,
                 microsecond=654320, tzinfo=MinutesFromUTC(120)),
        None,
        20,
        120,
        '20180912123040.65432*+120'
    ),
    (
        '20180912123040.6543**+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=40,
                 microsecond=654300, tzinfo=MinutesFromUTC(120)),
        None,
        19,
        120,
        '20180912123040.6543**+120'
    ),
    (
        '20180912123040.654***+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=40,
                 microsecond=654000, tzinfo=MinutesFromUTC(120)),
        None,
        18,
        120,
        '20180912123040.654***+120'
    ),
    (
        '20180912123040.65****+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=40,
                 microsecond=650000, tzinfo=MinutesFromUTC(120)),
        None,
        17,
        120,
        '20180912123040.65****+120'
    ),
    (
        '20180912123040.6*****+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=40,
                 microsecond=600000, tzinfo=MinutesFromUTC(120)),
        None,
        16,
        120,
        '20180912123040.6*****+120'
    ),
    (
        '20180912123040.******+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=40,
                 microsecond=0, tzinfo=MinutesFromUTC(120)),
        None,
        15,
        120,
        '20180912123040.******+120'
    ),
    (
        '2018091212304*.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '201809121230**.******+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=30, second=0,
                 microsecond=0, tzinfo=MinutesFromUTC(120)),
        None,
        12,
        120,
        '201809121230**.******+120'
    ),
    (
        '20180912123***.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '2018091212****.******+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=12, minute=0, second=0,
                 microsecond=0, tzinfo=MinutesFromUTC(120)),
        None,
        10,
        120,
        '2018091212****.******+120'
    ),
    (
        '201809121*****.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20180912******.******+120',
        'timestamp',
        datetime(year=2018, month=9, day=12, hour=0, minute=0, second=0,
                 microsecond=0, tzinfo=MinutesFromUTC(120)),
        None,
        8,
        120,
        '20180912******.******+120'
    ),
    (
        '2018091*******.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '201809********.******+120',
        'timestamp',
        datetime(year=2018, month=9, day=1, hour=0, minute=0, second=0,
                 microsecond=0, tzinfo=MinutesFromUTC(120)),
        None,
        6,
        120,
        '201809********.******+120'
    ),
    (
        '20180*********.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '2018**********.******+120',
        'timestamp',
        datetime(year=2018, month=1, day=1, hour=0, minute=0, second=0,
                 microsecond=0, tzinfo=MinutesFromUTC(120)),
        None,
        4,
        120,
        '2018**********.******+120'
    ),
    (
        '201***********.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20************.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '2*************.******+120',
        ValueError, None, None, None, None, None
    ),
    (    # year out of range for timestamp()
        '**************.******+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20140924193040.654321*120',
        ValueError, None, None, None, None, None
    ),
    (
        '20140924193040,654321+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20141324193040.654321+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20140932193040.654321+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20140924253040.654321+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20140924196140.654321+120',
        ValueError, None, None, None, None, None
    ),
    (
        '20140924193061.654321+120',
        ValueError, None, None, None, None, None
    ),
    (
        42,
        TypeError, None, None, None, None, None
    ),
], scope='module')
def datetime_init_tuple(request):
    """pytest.fixture that returns datatetime tuple info"""
    return request.param


def test_datetime_class_attrs_class():  # pylint: disable=invalid-name
    """Test class attrs via class level"""
    assert CIMDateTime.cimtype == 'datetime'


def test_datetime_class_attrs_inst():
    """Test class attrs via instance level"""
    obj = CIMDateTime('00000000000000.000000:000')
    assert obj.cimtype == 'datetime'


def test_datetime_inheritance():
    """Test inheritance"""
    obj = CIMDateTime('00000000000000.000000:000')
    assert isinstance(obj, CIMDateTime)
    assert isinstance(obj, CIMType)
    assert not isinstance(obj, CIMFloat)
    assert not isinstance(obj, CIMInt)


def test_datetime_init(datetime_init_tuple):
    """Test initialization from all input types using
       datetime_init_tuple pytest.fixture.
    """
    # pylint: disable=redefined-outer-name
    (dtarg, exp_kind, exp_datetime, exp_timedelta, exp_precision,
     exp_minutesfromutc, exp_str) = datetime_init_tuple
    try:
        obj = CIMDateTime(dtarg)
    except Exception as exc:  # pylint: disable=broad-except
        assert isinstance(exc, exp_kind)
    else:
        assert obj.is_interval == (exp_kind == 'interval')
        assert obj.datetime == exp_datetime
        if obj.datetime is not None:
            assert isinstance(obj.datetime, datetime)
            # We ensure that the datetime is always timezone-aware:
            assert obj.datetime.tzinfo is not None
        assert obj.timedelta == exp_timedelta
        if obj.timedelta is not None:
            assert isinstance(obj.timedelta, timedelta)
        assert obj.precision == exp_precision
        assert obj.minutes_from_utc == exp_minutesfromutc
        assert str(obj) == exp_str


def test_datetime_repr(datetime_init_tuple):
    """
    Test repr(CIMDateTime) from all input types using
    datetime_init_tuple pytest.fixture.
    """
    # pylint: disable=redefined-outer-name
    (dtarg, exp_kind, exp_datetime, exp_timedelta, exp_precision,
     exp_minutesfromutc, exp_str) = datetime_init_tuple

    if isinstance(exp_kind, type) and issubclass(exp_kind, Exception):
        pytest.skip("Testing repr() needs CIMDatetime object")

    obj = CIMDateTime(dtarg)

    # The code to be tested.
    # The actual test is that no exception is raised.
    repr(obj)


def test_datetime_str(datetime_init_tuple):
    """
    Test str(CIMDateTime) from all input types using
    datetime_init_tuple pytest.fixture.
    """
    # pylint: disable=redefined-outer-name
    (dtarg, exp_kind, exp_datetime, exp_timedelta, exp_precision,
     exp_minutesfromutc, exp_str) = datetime_init_tuple

    if isinstance(exp_kind, type) and issubclass(exp_kind, Exception):
        pytest.skip("Testing str() needs CIMDatetime object")

    obj = CIMDateTime(dtarg)

    # The code to be tested.
    # The actual test is that no exception is raised.
    str(obj)


# TODO: Add testcases for get_local_utcoffset()
# TODO: Add testcases for now()
# TODO: Add testcases for fromtimestamp()


TESTCASES_CIMTYPE = [

    # Testcases for cimtype()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: object to be tested.
    #   * exp_type_name: Expected CIM data type name.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger


    # Special cases
    (
        "Object is None",
        dict(
            obj=None,
            exp_type_name=None,
        ),
        TypeError, None, True
    ),

    # Boolean tests
    (
        "Object is a bool",
        dict(
            obj=True,
            exp_type_name=u'boolean',
        ),
        None, None, True
    ),

    # String tests
    (
        "Object is a unicode string",
        dict(
            obj=u"abc",
            exp_type_name=u'string',
        ),
        None, None, True
    ),
    (
        "Object is a byte string",
        dict(
            obj=b"abc",
            exp_type_name=u'string',
        ),
        None, None, True
    ),

    # Integer tests
    (
        "Object is an integer",
        dict(
            obj=42,
            exp_type_name=None,
        ),
        TypeError, None, True
    ),
    (
        "Object is a Uint8 number",
        dict(
            obj=Uint8(42),
            exp_type_name=u'uint8',
        ),
        None, None, True
    ),
    (
        "Object is a Uint16 number",
        dict(
            obj=Uint16(42),
            exp_type_name=u'uint16',
        ),
        None, None, True
    ),
    (
        "Object is a Uint32 number",
        dict(
            obj=Uint32(42),
            exp_type_name=u'uint32',
        ),
        None, None, True
    ),
    (
        "Object is a Uint64 number",
        dict(
            obj=Uint64(42),
            exp_type_name=u'uint64',
        ),
        None, None, True
    ),
    (
        "Object is a Sint8 number",
        dict(
            obj=Sint8(42),
            exp_type_name=u'sint8',
        ),
        None, None, True
    ),
    (
        "Object is a Sint16 number",
        dict(
            obj=Sint16(42),
            exp_type_name=u'sint16',
        ),
        None, None, True
    ),
    (
        "Object is a Sint32 number",
        dict(
            obj=Sint32(42),
            exp_type_name=u'sint32',
        ),
        None, None, True
    ),
    (
        "Object is a Sint64 number",
        dict(
            obj=Sint64(42),
            exp_type_name=u'sint64',
        ),
        None, None, True
    ),

    # Floating point tests
    (
        "Object is a float",
        dict(
            obj=42.0,
            exp_type_name=None,
        ),
        TypeError, None, True
    ),
    (
        "Object is a Real32 number",
        dict(
            obj=Real32(42.0),
            exp_type_name=u'real32',
        ),
        None, None, True
    ),
    (
        "Object is a Real64 number",
        dict(
            obj=Real64(42.0),
            exp_type_name=u'real64',
        ),
        None, None, True
    ),

    # Datetime tests
    (
        "Object is a Python datetime object",
        dict(
            obj=datetime(year=2014, month=9, day=24, hour=19, minute=30,
                         second=40, microsecond=654321,
                         tzinfo=MinutesFromUTC(120)),
            exp_type_name=u'datetime',
        ),
        None, None, True
    ),
    (
        "Object is a Python timedelta object",
        dict(
            obj=timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                          microseconds=654321),
            exp_type_name=u'datetime',
        ),
        None, None, True
    ),
    (
        "Object is a CIMDateTime object",
        dict(
            obj=CIMDateTime('20140924193040.654321+000'),
            exp_type_name=u'datetime',
        ),
        None, None, True
    ),

    # Other CIM object tests
    (
        "Object is a CIMClass object",
        dict(
            obj=CIMClass('CIM_Foo'),
            exp_type_name=u'string',  # embedded object
        ),
        None, None, True
    ),
    (
        "Object is a CIMInstance object",
        dict(
            obj=CIMInstance('CIM_Foo'),
            exp_type_name=u'string',  # embedded object
        ),
        None, None, True
    ),
    (
        "Object is a CIMInstanceName object",
        dict(
            obj=CIMInstanceName('CIM_Foo'),
            exp_type_name=u'reference',
        ),
        None, None, True
    ),
    (
        "Object is a CIMClassName object",
        dict(
            obj=CIMClassName('CIM_Foo'),
            exp_type_name=None,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_CIMTYPE)
@simplified_test_function
def test_cimtype(testcase, obj, exp_type_name):
    """
    Test function for cimtype().
    """

    # The code to be tested
    type_name = cimtype(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert type_name == exp_type_name


TESTCASES_TYPE_FROM_NAME = [

    # Testcases for type_from_name()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * type_name: CIM type name to be tested.
    #   * exp_type: Expected Python type object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Special cases
    (
        "Type name is None",
        dict(
            type_name=None,
            exp_type_obj=None,
        ),
        ValueError, None, True
    ),
    (
        "Type name is invalid",
        dict(
            type_name='foo',
            exp_type_obj=None,
        ),
        ValueError, None, True
    ),

    # CIM type names
    (
        "Type name is boolean",
        dict(
            type_name='boolean',
            exp_type_obj=bool,
        ),
        None, None, True
    ),
    (
        "Type name is uint8",
        dict(
            type_name='uint8',
            exp_type_obj=Uint8,
        ),
        None, None, True
    ),
    (
        "Type name is uint16",
        dict(
            type_name='uint16',
            exp_type_obj=Uint16,
        ),
        None, None, True
    ),
    (
        "Type name is uint32",
        dict(
            type_name='uint32',
            exp_type_obj=Uint32,
        ),
        None, None, True
    ),
    (
        "Type name is uint64",
        dict(
            type_name='uint64',
            exp_type_obj=Uint64,
        ),
        None, None, True
    ),
    (
        "Type name is sint8",
        dict(
            type_name='sint8',
            exp_type_obj=Sint8,
        ),
        None, None, True
    ),
    (
        "Type name is sint16",
        dict(
            type_name='sint16',
            exp_type_obj=Sint16,
        ),
        None, None, True
    ),
    (
        "Type name is sint32",
        dict(
            type_name='sint32',
            exp_type_obj=Sint32,
        ),
        None, None, True
    ),
    (
        "Type name is sint64",
        dict(
            type_name='sint64',
            exp_type_obj=Sint64,
        ),
        None, None, True
    ),
    (
        "Type name is real32",
        dict(
            type_name='real32',
            exp_type_obj=Real32,
        ),
        None, None, True
    ),
    (
        "Type name is real64",
        dict(
            type_name='real64',
            exp_type_obj=Real64,
        ),
        None, None, True
    ),
    (
        "Type name is string",
        dict(
            type_name='string',
            exp_type_obj=six.text_type,
        ),
        None, None, True
    ),
    (
        "Type name is char16",
        dict(
            type_name='char16',
            exp_type_obj=six.text_type,
        ),
        None, None, True
    ),
    (
        "Type name is datetime",
        dict(
            type_name='datetime',
            exp_type_obj=CIMDateTime,
        ),
        None, None, True
    ),
    (
        "Type name is reference",
        dict(
            type_name='reference',
            exp_type_obj=CIMInstanceName,
        ),
        None, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_TYPE_FROM_NAME)
@simplified_test_function
def test_type_from_name(testcase, type_name, exp_type_obj):
    """
    Test function for type_from_name().
    """

    # The code to be tested
    type_obj = type_from_name(type_name)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert type_obj is exp_type_obj


# TODO: Add tests for atomic_to_cim_xml()



"""
Wbemcli scriptlet displays the number of subscriptions, filters, and
distination objects on the defined server and the list of instance paths.
"""

from pywbem import WBEMServer, WBEMSubscriptionManager


def display_paths(instances, type_str):
    """Display the count and paths for the list of instances in instances."""
    print('%ss: count=%s' % (type_str, len(instances),))
    for path in [instance.path for instance in instances]:
        print('%s: %s' % (type_str, path))
    if len(instances):
        print('')


MY_SERVER = WBEMServer(CONN)  # noqa: F821
print('Interop namespace=%s' % MY_SERVER.interop_ns)

with WBEMSubscriptionManager('fred') as sub_mgr:
    server_id = sub_mgr.add_server(MY_SERVER)

    display_paths(sub_mgr.get_all_filters(server_id), 'Filter')

    display_paths(sub_mgr.get_all_destinations(server_id), 'Destination')

    display_paths(sub_mgr.get_all_subscriptions(server_id), 'Subscription')

print('Quit Scriplet. Quiting wbemcli')
quit()
